
// Name
//===================================
//
// Sprite
//


// DEPENDENCIES
//===================================

    #include "block.h"
    #include "camera.h"
    #include "collision.h"
    #include "sprite.h"
    #include "sprite_graphics.h"
    #include "sprite_movement.h"
    #include "input.h"
    #include "unit.h"


// STATIC PROPERTIES
//===================================

// METHODS
//===================================

    Sprite::Sprite
    (
        SpriteGraphics* graphics,
        SpriteMovement* movement,
        int x,
        int y,
        int width,
        int height,
        Type type,
        int texture_x,
        int texture_y
    )
    :
        Object( x, y, width, height ),
        graphics_ ( graphics ),
        movement_ ( movement ),
        type_ ( type ),
        texture_x_ ( texture_x ),
        texture_y_ ( texture_y )
    {};

    Sprite::~Sprite()
    {
        delete movement_;
        delete graphics_;
    };

    void Sprite::update( Input& input, Camera& camera, Map& lvmap, Game& game, EventSystem& events )
    {
        if ( movement_ != nullptr )
            movement_->update( *this, input, camera, lvmap, graphics_, game, events );

        position();

        if ( graphics_ != nullptr )
        {
            graphics_->update( *this );
        }
    };

    void Sprite::render( Graphics& graphics, Camera& camera )
    {
        if ( graphics_ != nullptr )
        {
            graphics_->render( graphics, adjustBoundBoxes(), &camera );
        }
    };

    void Sprite::moveLeft()
    {
        acceleration_x_ = -start_speed_;
        direction_x_ = Direction::Horizontal::LEFT;
    };

    void Sprite::moveRight()
    {
        acceleration_x_ = start_speed_;
        direction_x_ = Direction::Horizontal::RIGHT;
    };

    void Sprite::stop()
    {
        acceleration_x_ = 0;
    };

    void Sprite::stopDucking()
    {
        acceleration_x_ = 0;
    };

    void Sprite::position()
    {
        x_prev_ = physics_bound_box_.x;
        y_prev_ = physics_bound_box_.y;

        if ( isJumping() )
        {
            acceleration_x_ *= .75;
        }

        if ( !onGround() )
            top_speed_ = TOP_SPEED_WALK * .95;

        if ( isDucking() )
        {
            acceleration_x_ /= 3;
            jump_top_speed_ = JUMP_DUCK_TOP_SPEED;
        }

        vx_ += acceleration_x_;

        if ( acceleration_x_ == 0 )
            vx_ /= 1.1;

        if ( vx_ > top_speed_ )
            vx_ = top_speed_;
        if ( vx_ < -top_speed_ )
            vx_ = -top_speed_;

        physics_bound_box_.x += vx_;

        top_speed_ = TOP_SPEED_WALK;
        start_speed_ = START_SPEED_WALK;

        is_running_ = false;


        // VERTICAL MOVEMENT

        if ( abs( top_speed_ ) > TOP_SPEED_WALK )
            jump_top_speed_ *= 1.5;

        if ( !onGround() && !isJumping() && !onLadder() )
        {
            acceleration_y_ = gravity_start_speed_;
        }

        vy_ += acceleration_y_;

        if ( vy_ > gravity_top_speed_ )
            vy_ = gravity_top_speed_;

        if ( vy_ < -jump_top_speed_ )
        {
            vy_ = -jump_top_speed_;
            jump_start_ = false;
        }

        physics_bound_box_.y += vy_;

        gravity_start_speed_ = GRAVITY_START_SPEED;
        gravity_top_speed_ = GRAVITY_TOP_SPEED;
        jump_top_speed_ = JUMP_TOP_SPEED;

        if ( acceleration_x_ != 0 || acceleration_y_ != 0 )
            is_moving_ = true;
        else
            is_moving_ = false;

        acceleration_y_ = 0;

        if ( !isJumping() )
            jump_end_ = true;

        is_jumping_ = false;
        on_ground_prev_ = on_ground_;
        on_ground_ = false;

        if ( onLadder() )
        {
            vy_ = 0;
        }
    };

    void Sprite::boundaries( Camera& camera )
    {
        if ( physics_bound_box_.x < camera.x() )
        {
            collideStopXLeft( abs ( camera.x() - physics_bound_box_.x ) );
        }
        if ( physics_bound_box_.x + physics_bound_box_.w > Unit::PixelsToSubPixels( camera.right() ) )
        {
            collideStopXRight( ( physics_bound_box_.x + physics_bound_box_.w ) - Unit::PixelsToSubPixels( camera.right() ) );
        }
    };

    void Sprite::collideStopXLeft( int overlap )
    {
        stop();

        if ( vx_ < 0 )
        {
            physics_bound_box_.x -= vx_;
            vx_ = -(vx_ * BOUNCE);
        }
    };

    void Sprite::collideStopXRight( int overlap )
    {
        stop();

        if ( vx_ > 0 )
        {
            physics_bound_box_.x -= vx_;
            vx_ = -(vx_ * BOUNCE);
        }
    };

    void Sprite::collideStopYBottom( int overlap )
    {
        if ( vy_ > 0 )
        {
            acceleration_y_ = 0;

            if ( !onGround() )
            {
                physics_bound_box_.y -= overlap;
            }

            on_ground_ = true;
        }

        if ( onLadder() )
        {
            //std::cout<<"LAD"<<std::endl;
            physics_bound_box_.y -= 1000;
        }
    };

    void Sprite::collideStopYTop( int overlap )
    {
        acceleration_y_ = 0;
        physics_bound_box_.y += vy_;
        vy_ = 0;
        physics_bound_box_.y += overlap;
        jump_end_ = true;
    };

    bool Sprite::onGround() const
    {
        return on_ground_;
    };

    bool Sprite::isJumping() const
    {
        return is_jumping_;
    };

    void Sprite::jump()
    {
        if ( !onGround() || !jump_lock_ )
        {
            if ( onGround() && !jump_lock_ )
            {
                startJump();
            }

            if ( jump_start_ && !jump_end_ )
            {
                is_jumping_ = true;
            }

            if ( isJumping() )
            {
                acceleration_y_ = -JUMP_START_SPEED;
            }
        }
    };

    void Sprite::startJump()
    {
        jump_start_ = true;
        jump_end_ = false;
        vy_ = 0;
        on_ground_ = false;
    };

    void Sprite::slowFall()
    {
        gravity_start_speed_ = GRAVITY_START_SPEED * .75;
        gravity_top_speed_ = GRAVITY_TOP_SPEED * .85;
    };

    bool Sprite::isRunning() const
    {
        return is_running_;
    };

    void Sprite::run()
    {
        is_running_ = true;
        top_speed_ = TOP_SPEED_RUN;
        start_speed_ = START_SPEED_RUN;
    };

    SDL_Rect Sprite::adjustBoundBoxes()
    {

        int h = heightPixels() + 1;
        int y = yPixels() - 1;

        if ( isDucking() )
        {
            h += 4;
            y -= 4;
        }

        return
        {
            xPixels() - 1,
            y,
            width_,
            h
        };
    };

    void Sprite::collisionDetect( Sprite& other )
    {
        if ( movement_ != nullptr )
            movement_->collision( collide( &other ), *this, other );
    };

    void Sprite::collisionDetect( Block& block, Level& level, EventSystem& events, Inventory& inventory, Camera& camera, GameState* state, Graphics& graphics )
    {
        if ( movement_ != nullptr )
            movement_->collision( collide( &block ), *this, block, level, events, inventory, camera, state, graphics );
    };

    bool Sprite::canJump() const
    {
        return can_jump_;
    };

    void Sprite::debugResetPos()
    {
        physics_bound_box_.x = 96000;
        physics_bound_box_.y = 64000;
    };

    bool Sprite::isDucking() const
    {
        return is_ducking_;
    };

    void Sprite::duck()
    {
        // Can continue ducking while in air, but only start duck on ground.
        if ( isDucking() || onGround() )
        {
            // Hacky way to make player warp to the right position after height changes.
            if ( !isDucking() )
                physics_bound_box_.y += 10000;

            is_ducking_ = true;

            physics_bound_box_.h = Unit::PixelsToSubPixels( height_ - 11 );
        }
    };

    void Sprite::unduck()
    {
        // Hacky way to keep player from falling through ground after gaining height from unducking.
        if ( isDucking() )
            physics_bound_box_.y -= 10000;

        is_ducking_ = false;
        physics_bound_box_.h = Unit::PixelsToSubPixels( height_ - 1 );
    };

    Direction::Horizontal Sprite::directionX() const
    {
        return direction_x_;
    };

    bool Sprite::onGroundPrev() const
    {
        return on_ground_prev_;
    };

    bool Sprite::isMoving() const
    {
        return is_moving_;
    };

    void Sprite::stopY()
    {
        acceleration_y_ = 0;
        vy_ = 0;
    };

    Sprite::Type Sprite::type() const
    {
        return type_;
    };

    bool Sprite::onLadder() const
    {
        return on_ladder_;
    };

    bool Sprite::onLadderPrev() const
    {
        return on_ladder_prev_;
    };

    bool Sprite::touchingLadder() const
    {
        return touching_ladder_;
    };

    bool Sprite::touchingLadderPrev() const
    {
        return touching_ladder_prev_;
    };

    void Sprite::touchLadder()
    {
        touching_ladder_ = true;
    };

    void Sprite::grabLadder()
    {
        if ( !isDucking() )
            on_ladder_ = true;
    };

    void Sprite::releaseLadder()
    {
        on_ladder_ = false;
    };

    void Sprite::moveUp()
    {
        if ( onLadder() )
        {
            acceleration_y_ = -1000;
        }
        else if ( onGround() )
        {
            lookUp();
        }
    };

    void Sprite::moveDown()
    {
        if ( onLadder() )
        {
            acceleration_y_ = +1000;
        }
    };

    void Sprite::lookUp()
    {
        looking_up_ = true;
    };

    bool Sprite::lookingUp() const
    {
        return looking_up_;
    };

    void Sprite::stopLookingUp()
    {
        looking_up_ = false;
    };

    void Sprite::kill()
    {
        dead_ = true;
    };

    bool Sprite::isDead() const
    {
        return dead_;
    };

    void Sprite::setPosition( int x, int y )
    {
        physics_bound_box_.x = Unit::PixelsToSubPixels( x );
        physics_bound_box_.y = Unit::PixelsToSubPixels( y );
    };
